##	LexisNexis.filter.Feb14.py
##
##  This takes sentences of stories from an assortment of Canadian media sources 
##  downloaded from Lexis-Lexis for the Dugan-Chenoweth and puts these into TABARI
##  format. The program recognizes the sources in the sourceids dictionaries below:
##  it seems fairly robust and probably all that will be required to add most sources
##  is simply finding the text that follows the "NNN of MMMM DOCUMENTS" documents line
##  and adding this plus an abbreviation to the dictionary. It seems to generalize fairly
##  well to current LN formats.
##
##	TO RUN PROGRAM:
##
##	python LexisNexis.filter.Feb14.py <command line options>
##
##	COMMAND LINE OPTIONS:
##
##  -i <filename>: process only <filename> rather than the complete list
##
##  -l <filename>: get the list of files from <filename> rather than "format.files.txt"
##
##  -o <filename>: output file name; default is "filter.outfile.txt"
##
##  Example 1
##    python LexisNexis.filter.Feb13.py
##    (reads the list of input files from "files.list.txt", the default file list, and 
##	   writes the draft output to "filter.outfile.txt", the default output file name)
##
##  Example 2
##    python LexisNexis.filter.Feb13.py -l levant.list.txt -o levant.formatted.txt
##    (reads the list of input files from "levant.list.txt" and writes the output to
##	   "levant.formatted.txt")
##
##  Example 3
##    python LexisNexis.filter.Feb13.py -i factiva.test.html -d
##    (reads the single input file from "factiva.test.html" and writes the draft output to
##	   "filter.outfile.txt", the default output file name)
##
##	INPUT FILES
##
##  format.files.txt: list of the files to process. The last line in the file should be
##                  blank (that is, only a return \n)
##  
##	OUTPUT FILES
##
##  Program TABARI-formated records with the following format:
##    YYMMDD AAA-MMMM-NN filename
##  where
##    YYMMDD: date
##    AAA:    source abbreviation from sourceids{}
##    MMMM:   zero-filled sequence of the story within the source-day
##    NN:     sentence number within the story
## 
##  Text is wrapped to 80 columns
##
##	PROGRAMMING NOTES: 
##
##  1. See notes in sentence_segmenter() on the conditions for segmenting sentences. The
##     list of abbreviations can be extended in the global ABBREV_LIST. Also note the
##     length restrictions in MIN/MAX_SENTLENGTH.
##
##     This uses a custom segmenter rather than the nltk "punkt" routine since this has
##     less overhead and also is based on explicit patterns rather than trained. That said,
##     punkt may work better in some applications and such easily be substituted. 
##
##  2. Normal exit will print "Finished" Unexpected EOF conditions generates a message
##     "Exiting on EOF at mark N" where N is unique in the program.
##
##  3. The list of sources in the dictionary sourceids is required only if you want to
##     assign source abbreviations; it could be replaced with a regex that looks for
##     "\d+ of \d+ DOCUMENTS" and then goes to the second long line to get the date.
##
##  4. This works with both Unix and Windows line endings.
##
##  5. The sentence segmenter could still use a bit of work on how it handles quotes --
##     for example it is not splitting on `."'
##
##	SYSTEM REQUIREMENTS
##	This program has been successfully run under Mac OS 10.6; it is standard Python 2.5
##	so it should also run in Unix or Windows. 
##
##	PROVENANCE:
##	Programmer: Philip A. Schrodt
##				Parus Analytical Systems
##				schrodt735@gmail.com
##				http://eventdata.parsuanalytics.edu
##
##	Copyright (c) 2014	Philip A. Schrodt.	All rights reserved.
##
##	Redistribution and use in source and binary forms, with or without modification,
##	are permitted under the terms of the GNU General Public License:
##	http://www.opensource.org/licenses/gpl-license.html
##
##	Report bugs to: schrodt735@gmail.com
##
##	REVISION HISTORY:
##	16-Feb-14:	Initial version written for Dugan-Chenoweth filtering
##
##	----------------------------------------------------------------------------------

import sys
import textwrap
import time
import re

# ======== global initializations ========= #

sourceids = {
    'Canadian Government News': 'CGN',
    'The Financial Post': 'FPO',
    'Edmonton Journal (Alberta)': 'EDJ',
    'The Toronto Star': 'TST',
    'The Gazette (Montreal)': 'GAZ',
    'The Gazette (Montreal, Quebec)': 'GAZ',
    'The Halifax Daily News (Nova Scotia)': 'HDN',
    'Agence France Presse -- English': 'AFP',
}


headerflags = ['LENGTH', 'SECTION', 'BYLINE', 'DATELINE']  # various indicators we are past the headline

datepat = re.compile(',? \w+day')	# used to isolate '%B %d, %Y' across various date line formats
terpat = re.compile('[\.\?!]\s+[A-Z\"]')	# sentence termination pattern used in sentence_segmenter(paragr)

MIN_SENTLENGTH = 80   # this is relatively high because we are only looking for sentences that will have subject and object
MAX_SENTLENGTH = 512
STORY_LINE_SIZE = 64  # first line following a header greater than this length is start of story

#source: LbjNerTagger1.11.release/Data/KnownLists/known_title.lst from University of Illinois with editing
ABBREV_LIST = ['mrs.', 'ms.', 'mr.', 'dr.', 'gov.', 'sr.', 'rev.',  'r.n.', 'pres.', 
	'treas.', 'sect.', 'maj.', 'ph.d.', 'ed. psy.', 'proc.', 'fr.', 'asst.', 'p.f.c.', 'prof.', 
	'admr.', 'engr.', 'mgr.', 'supt.', 'admin.', 'assoc.', 'voc.', 'hon.', 'm.d.', 'dpty.', 
	'sec.', 'capt.', 'c.e.o.', 'c.f.o.', 'c.i.o.', 'c.o.o.', 'c.p.a.', 'c.n.a.', 'acct.', 
	'llc.', 'inc.', 'dir.', 'esq.', 'lt.', 'd.d.', 'ed.', 'revd.', 'psy.d.', 'v.p.', 
	'senr.', 'gen.', 'prov.', 'cmdr.', 'sgt.', 'sen.', 'col.', 'lieut.', 'cpl.', 'pfc.', 
	'k.p.h.', 'cent.', 'deg.', 'doz.', 'Fahr.', 'Cel.', 'F.', 'C.', 'K.', 'ft.', 'fur.', 
	'gal.', 'gr.', 'in.', 'kg.', 'km.', 'kw.', 'l.', 'lat.', 'lb.', 'lb per sq in.', 
	'long.', 'mg.', 'mm.,, m.p.g.', 'm.p.h.', 'cc.', 'qr.', 'qt.', 'sq.', 't.', 'vol.', 
	'w.', 'wt.']
	
storycount = {}   # number of stories in current day
sourceday = {}    # current day for the source

dirfile = "format.files.txt"   	# default file list
outfile = "filter.outfile.txt"  # default output
dosingle = False

wrapper = textwrap.TextWrapper(width = 79)

# ============ function definitions ================ #

def do_exit(mark):
	print 'Exiting on EOF at mark',mark
	sys.exit()
	
def sentence_segmenter(paragr):
	""" Breaks the string 'paragraph' into a list of sentences based on the following rules
	1. Look for terminal [.,?,!] followed by a space and [A-Z]
	2. If ., check against abbreviation list ABBREV_LIST: Get the string between the . and the
	   previous blank, lower-case it, and see if it is in the list. Also check for single-
	   letter initials. If true, continue search for terminal punctuation
	3. Extend selection to balance (...) and "...". Reapply termination rules
	4. Add to sentlist if the length of the string is between MIN_SENTLENGTH and MAX_SENTLENGTH 
	5. Returns sentlist """

#	ka = 0
#	print '\nSentSeg-Mk1'
	sentlist = []
	searchstart = 0  # controls skipping over non-terminal conditions
	terloc = terpat.search(paragr)
	while terloc:
#		print 'Mk2-0:', paragr[:terloc.start()+2]
		isok = True
		if paragr[terloc.start()] == '.':
			if (paragr[terloc.start()-1].isupper() and 
				paragr[terloc.start()-2] == ' '):  isok = False   # single initials
			else:
				loc = paragr.rfind(' ',0,terloc.start()-1)   # check abbreviations
				if loc > 0:
#					print 'SentSeg-Mk1: checking',paragr[loc+1:terloc.start()+1]
					if paragr[loc+1:terloc.start()+1].lower() in ABBREV_LIST: 
#						print 'SentSeg-Mk2: found',paragr[loc+1:terloc.start()+1]
						isok = False
		if paragr[:terloc.start()].count('(') != paragr[:terloc.start()].count(')') :  
#			print 'SentSeg-Mk2: unbalanced ()'
			isok = False
		if paragr[:terloc.start()].count('"') % 2 != 0  :
#			print 'SentSeg-Mk2: unbalanced ""'
			isok = False
		if isok:
			if (len(paragr[:terloc.start()]) > MIN_SENTLENGTH and   
				len(paragr[:terloc.start()]) < MAX_SENTLENGTH) :
				sentlist.append(paragr[:terloc.start()+2])
#				print 'SentSeg-Mk3: added',paragr[:terloc.start()+2]
			paragr = paragr[terloc.end()-1:]
			searchstart = 0
		else: searchstart = terloc.start()+2 
 
#		print 'SentSeg-Mk4:',paragr[:64]
#		print '            ',paragr[searchstart:searchstart+64]
		terloc = terpat.search(paragr,searchstart)
#		ka += 1
#		if ka > 16: sys.exit()

	if (len(paragr) > MIN_SENTLENGTH and len(paragr) < MAX_SENTLENGTH) :  # add final sentence 
		sentlist.append(paragr) 
		
	return sentlist

def write_records():  
	nsent = 1
	for sent in sentlist:
		fout.write(datestr + ' ' + source + '-' + str(storycount[source]).zfill(4) + '-' + str(nsent).zfill(2) + '\t' + infile[:-1] + '\n')
		lines = textwrap.wrap(sent, 80)
		for txt in lines: fout.write(txt+' \n')
#				print txt
		fout.write('\n')
		nsent += 1

# ============ main program =============== #

if len(sys.argv) > 1:   # process command line options
	if ("-l" in sys.argv):
		dirfile = sys.argv[sys.argv.index("-l")+1]	
		print "-l: Reading file list from",dirfile
	if ("-i" in sys.argv):
		dosingle = True
		infile = sys.argv[sys.argv.index("-i")+1] + " " # the terminal space gets removed	
		print "-i: Processing single file",infile[:-1]
	if ("-o" in sys.argv):
		outfile = sys.argv[sys.argv.index("-o")+1]	
		print "-o: Output file",outfile

fout = open(outfile,'w')

if not dosingle: 
	try: 
		fdir = open(dirfile,'r')
	except IOError:
		print "\aError: Could not find the file list file",dirfile
		sys.exit()	
	infile = fdir.readline() 

while len(infile) > 8:  # loop through the file list; stop at blank line

	try: 
		fin = open(infile[:-1],'r')
	except IOError:
		print "\aError: Could not find the file",infile[:-1]
		sys.exit()	
	print "\nReading file:",infile,

	caseno = 1
	noarticle = True
	line = fin.readline()
	while len(line) > 0:  # loop through the file
	
		if line.strip() not in sourceids: 
			line = fin.readline()
			continue
		source = sourceids[line.strip()]
											# get the date line
		caseno += 1
		if caseno > 16: break
		line = fin.readline()
		dateseg = '---'
		while len(line) < 8: line = fin.readline()
		line = line.strip()
		try: 
			if source == 'AFP': dateseg = line
			else: 
				mat = datepat.search(line)
				dateseg = line[:mat.start()]
			strctim = time.strptime(dateseg,'%B %d, %Y')
			datestr = time.strftime('%y%m%d',strctim)
#			print source, dateseg, datestr
		except ValueError:
			print 'Cannot interpret date',dateseg
			line = fin.readline()
			continue	
		
		noarticle = False
		if source in sourceday and sourceday[source] == datestr:  # update story count for source
			storycount[source] += 1 
		else:
			storycount[source] = 1 
			sourceday[source] = datestr  # reset date

		print source, dateseg, 'Story',storycount[source]

		line = fin.readline()  # get past the header
		while len(line) > 0:
			if ':' in line:
				if line[:line.index(':')] in headerflags: break 
			line = fin.readline()
		if len(line) == 0: do_exit(2)  # eof check
			
		while len(line) < STORY_LINE_SIZE and len(line) > 0: line = fin.readline()
		if len(line) == 0: do_exit(3)
		story = line[:-1]
			
		line = fin.readline()
		while (len(line) > 0 and 
				'LOAD-DATE' not in line.upper() and
				not line.upper().startswith('LANGUAGE:')):
			if len(line) > 2: 
				story += line[:-1]
				if story[-1] != ' ': story += ' '
			line = fin.readline()
		if len(line) == 0: do_exit(4)
		
		story = "".join(i for i in story if ord(i)<128 and ord(i) > 31)  # filter non-ASCII characters
		
#		lines = textwrap.wrap(story, 80)
#		print 'Story:'
#		for txt in lines: print txt 

		sentlist = sentence_segmenter(story)
				
		write_records()			
	
		line = fin.readline()
	
	fin.close()
	if noarticle :
		print "\aError: could not find any articles in the file",infile,
		print "Please check the file formating and re-run"
		print "Ending processing at this point"
		break 
	if dosingle: break  # normal termination when doing only one file 
	else: infile = fdir.readline() 

fout.close()

print "Finished"

""" Test text for the segmenter

Sealed Air (Canada) CO/CIE, a Nova Scotia manufacturer of packaging products, was fined
$55,000 today for a violation of the Occupational Health and Safety Act after a
worker was injured. On March 15, 2010, at the company's plant in Brampton, a temporary 
worker, Mr. John J. Jones was applying tape, sticky side out, to a cardboard roll that
was rotating on a machine. The worker placed a hand (on the ABM. Roll) and the
previously applied tape caught the worker's glove. The worker's arm was pulled
around the roller and fractured!  A Ministry of Labour (MIN. LAB.) investigation found "that
the MBA. Worker" had not been given specific information about the hazards of placing
a gloved hand on the cardboard roll.  Sealed Air (Canada) CO/CIE pleaded guilty
to failing to provide adequate information, instruction and supervision to the
worker.  Was the fine imposed by Justice Karen Jensnen? In addition to
the fine, the court imposed a 25-per-cent victim fine surcharge, as required by
the Provincial Offences Act. The surcharge is credited to a special provincial
government fund to assist victims of crime.

Output sentences:

Sealed Air (Canada) CO/CIE, a Nova Scotia manufacturer of packaging 
products, was fined $55,000 today for a violation of the Occupational Health and 
Safety Act after a worker was injured. 

On March 15, 2010, at the company's plant in Brampton, a temporary worker, Mr. 
John J. Jones was applying tape, sticky side out, to a cardboard roll that was 
rotating on a machine. 

The worker placed a hand (on the ABM. Roll) and the previously applied tape 
caught the worker's glove. 

The worker's arm was pulled around the roller and fractured! 

A Ministry of Labour (MIN. LAB.) investigation found "that the MBA. Worker" had 
not been given specific information about the hazards of placing a gloved hand 
on the cardboard roll. 

Sealed Air (Canada) CO/CIE pleaded guilty to failing to provide adequate 
information, instruction and supervision to the worker. 

Was the fine imposed by Justice Karen Jensnen? 

In addition to the fine, the court imposed a 25-per-cent victim fine surcharge, 
as required by the Provincial Offences Act. 

The surcharge is credited to a special provincial government fund to assist 
victims of crime. 
"""

